import math

def getDistanceFromCoord(lon1, lat1, lon2, lat2):
    # return km
    lon1, lat1, lon2, lat2 = map(math.radians, [lon1, lat1, lon2, lat2])
    dlon = lon2 - lon1
    dlat = lat2 - lat1
    a = math.sin(dlat / 2) ** 2 + math.cos(lat1) * math.cos(lat2) * math.sin(dlon / 2) ** 2
    c = 2 * math.asin(math.sqrt(a))
    r = 6371
    return c * r


osm_highway_type_dict = {'motorway': 'motorway',
                         'motorway_link': 'motorway',
                         'trunk': 'trunk',
                         'trunk_link': 'trunk',
                         'primary': 'primary',
                         'primary_link': 'primary',
                         'secondary': 'secondary',
                         'secondary_link': 'secondary',
                         'tertiary': 'tertiary',
                         'tertiary_link': 'tertiary',
                         'residential': 'residential',
                         'unclassified': 'unclassified'}

link_type_no_dict = {'motorway':1, 'trunk':2, 'primary':3, 'secondary':4, 'tertiary':5, 'residential':6, 'unclassified':7}

default_lanes_dict = {'motorway': 4, 'trunk': 3, 'primary': 3, 'secondary': 2, 'tertiary': 2, 'residential': 1, 'unclassified': 1}
default_speed_dict = {'motorway': 59, 'trunk': 39, 'primary': 39, 'secondary': 39, 'tertiary': 29, 'residential': 29, 'unclassified': 29}


default_oneway_flag_dict = {'motorway': True,'trunk':True,'primary':True,'secondary':False,'tertiary':False,'residential':False,'unclassified':False}


default_int_buffer = 20.0





from .settings import *
from .readfile import *
from .simplification import *
from .complex_intersection import *
from .wayfilters import *
import re


def parseNodes(network, nodes, strict_mode):
    for osm_node in nodes:
        node = Node()
        node.osm_node_id = int(osm_node.attrib['id'])
        node.node_no = network.created_nodes
        node.x_coord = float(osm_node.attrib['lon'])
        node.y_coord = float(osm_node.attrib['lat'])

        if strict_mode:
            if not ((network.minlon <= node.x_coord <= network.maxlon) and (network.minlat <= node.y_coord <= network.maxlat)):
                node.in_region = False

        for info in osm_node:
            if info.tag == 'tag':
                if info.attrib['k'] == 'highway':
                    node.osm_highway = info.attrib['v']

        if 'signal' in node.osm_highway: node.ctrl_type = 1         # todo: check signalized tag

        network.node_set.add(node)
        network.osm_node_id_to_node_dict[node.osm_node_id] = node
        network.created_nodes += 1


def newLinkFromWay(link_no, way, direction, ref_node_list):
    link = Link()
    link.osm_way_id = way.osm_way_id
    link.link_no = link_no
    link.name = way.name
    link.link_type_name = way.link_type_name
    link.link_type = way.link_type
    link.free_speed = way.maxspeed
    link.allowed_uses = way.allowed_uses

    if way.oneway:
        link.lanes_list = [way.lanes]
    else:
        if direction == 1:
            if way.forward_lanes is not None:
                link.lanes_list = [way.forward_lanes]
            elif way.lanes is not None:
                link.lanes_list = [math.ceil(way.lanes / 2)]
            else:
                link.lanes_list = [way.lanes]
        else:
            if way.backward_lanes is not None:
                link.lanes_list = [way.backward_lanes]
            elif way.lanes is not None:
                link.lanes_list = [math.ceil(way.lanes / 2)]
            else:
                link.lanes_list = [way.lanes]

    link.lanes = link.lanes_list[0]
    link.from_node = ref_node_list[0]
    link.to_node = ref_node_list[-1]
    link.from_node.outgoing_link_list.append(link)
    link.to_node.incoming_link_list.append(link)
    for ref_node in ref_node_list: link.geometry_point_list.append((ref_node.x_coord,ref_node.y_coord))
    link.getGeometryStr()
    link.calculateLength()

    return link


def getCoordOnBoundary(node_in,node_outside,network):
    if node_in.x_coord == node_outside.x_coord:
        x_coord = node_in.x_coord
        if node_outside.y_coord > network.maxlat:
            y_coord = network.maxlat
        else:
            y_coord = network.minlat
        return x_coord, y_coord

    if node_outside.x_coord > network.maxlon:
        x_coord = network.maxlon
        y_coord = (node_outside.y_coord - node_in.y_coord)/(node_outside.x_coord - node_in.x_coord) * (x_coord-node_in.x_coord) + node_in.y_coord
        if network.minlat <= y_coord <= network.maxlat:
            return x_coord, y_coord

    if node_outside.x_coord < network.minlon:
        x_coord = network.minlon
        y_coord = (node_outside.y_coord - node_in.y_coord)/(node_outside.x_coord - node_in.x_coord) * (x_coord-node_in.x_coord) + node_in.y_coord
        if network.minlat <= y_coord <= network.maxlat:
            return x_coord, y_coord

    if node_outside.y_coord > network.maxlat:
        y_coord = network.maxlat
        x_coord = (node_outside.x_coord - node_in.x_coord) / (node_outside.y_coord - node_in.y_coord) * (y_coord - node_in.y_coord) + node_in.x_coord
        if network.minlon <= x_coord <= network.maxlon:
            return x_coord, y_coord

    if node_outside.y_coord < network.minlat:
        y_coord = network.minlat
        x_coord = (node_outside.x_coord - node_in.x_coord) / (node_outside.y_coord - node_in.y_coord) * (y_coord - node_in.y_coord) + node_in.x_coord
        if network.minlon <= x_coord <= network.maxlon:
            return x_coord, y_coord

    raise Exception('cannot build a node on the boundary')



def createNodeOnBoundary(node_in,node_outside,network):
    node = Node()
    node.node_no = network.created_nodes
    node.x_coord, node.y_coord = getCoordOnBoundary(node_in,node_outside,network)
    node.is_crossing = True

    network.node_set.add(node)
    network.osm_node_id_to_node_dict[node.osm_node_id] = node
    network.created_nodes += 1

    return node


def getSegmentNodeList(way, segment_no, network):
    m_segment_node_list = way.segment_node_list[segment_no]
    if way.is_reversed: m_segment_node_list = list(reversed(m_segment_node_list))
    number_of_nodes = len(m_segment_node_list)

    m_segment_node_list_group = []

    if m_segment_node_list[0].in_region:
        idx_first_outside = -1
        for idx, node in enumerate(m_segment_node_list):
            if not node.in_region:
                idx_first_outside = idx
                break

        if idx_first_outside == -1:
            m_segment_node_list_group.append(m_segment_node_list)
            return m_segment_node_list_group
        else:
            new_node = createNodeOnBoundary(m_segment_node_list[idx_first_outside-1],m_segment_node_list[idx_first_outside], network)
            m_segment_node_list_group.append(m_segment_node_list[:idx_first_outside] + [new_node])

    if m_segment_node_list[-1].in_region:
        idx_last_outside = -1
        for idx in range(number_of_nodes-2,-1,-1):
            if not m_segment_node_list[idx].in_region:
                idx_last_outside = idx
                break
        new_node = createNodeOnBoundary(m_segment_node_list[idx_last_outside+1],m_segment_node_list[idx_last_outside], network)
        m_segment_node_list_group.append([new_node] + m_segment_node_list[idx_last_outside+1:])

    return m_segment_node_list_group


def createLinks(network, way_list):
    for way in way_list:
        if way.is_cycle:
            crossing_in_cycle = False
            for node in way.ref_node_list[1:-1]:
                if node.is_crossing:
                    crossing_in_cycle = True
                    break
            if not crossing_in_cycle: continue

        way.getNodeListForSegments()
        for segment_no in range(way.number_of_segments):

            m_segment_node_list_group = getSegmentNodeList(way, segment_no, network)
            for m_segment_node_list in m_segment_node_list_group:
                link = newLinkFromWay(network.created_links, way, 1, m_segment_node_list)
                network.link_set.add(link)
                network.created_links += 1
                if not way.oneway:
                    linkb = newLinkFromWay(network.created_links, way, -1, list(reversed(m_segment_node_list)))
                    network.link_set.add(linkb)
                    network.created_links += 1


            # m_segment_node_list = way.segment_node_list[segment_no]
            # valid_flag = True
            # for node in m_segment_node_list:
            #     if not node.in_region:
            #         valid_flag = False
            #         break
            # if not valid_flag: continue
            #
            # if way.is_reversed:
            #     link = newLinkFromWay(network.created_links, way, 1, list(reversed(m_segment_node_list)))
            #     network.link_set.add(link)
            #     network.created_links += 1
            # else:
            #     link = newLinkFromWay(network.created_links, way, 1, m_segment_node_list)
            #     network.link_set.add(link)
            #     network.created_links += 1
            #     if not way.oneway:
            #         linkb = newLinkFromWay(network.created_links, way, -1, list(reversed(m_segment_node_list)))
            #         network.link_set.add(linkb)
            #         network.created_links += 1


def parseWays(network, ways, network_type):
    way_list = []
    network_type_set = set(network_type)


    for osm_way in ways:
        way = Way()
        way.osm_way_id = int(osm_way.attrib['id'])

        for info in osm_way:
            if info.tag == 'nd':
                ref_node_id = int(info.attrib['ref'])
                try:
                    way.ref_node_list.append(network.osm_node_id_to_node_dict[ref_node_id])
                except KeyError:
                    print(f'  warning: ref node {ref_node_id} in way {way.osm_way_id} is not defined')          # todo: exception
                    return
            elif info.tag == 'tag':
                if info.attrib['k'] == 'highway':
                    way.highway = info.attrib['v']
                elif info.attrib['k'] == 'lanes':
                    try:
                        way.lanes = int(info.attrib['v'])
                    except ValueError:
                        way.lanes = int(info.attrib['v'].split(';')[0])
                elif info.attrib['k'] == 'lanes:forward':
                    way.forward_lanes = int(info.attrib['v'])
                elif info.attrib['k'] == 'lanes:backward':
                    way.backward_lanes = int(info.attrib['v'])
                elif info.attrib['k'] == 'name':
                    way.name = info.attrib['v']
                elif info.attrib['k'] == 'maxspeed':
                    way.maxspeed = re.findall(r'\d+\.?\d*', info.attrib['v'])[0]
                elif info.attrib['k'] == 'oneway':
                    oneway_flag = info.attrib['v']
                    if oneway_flag == 'yes' or oneway_flag == '1':
                        way.oneway = True
                    elif oneway_flag == 'no' or oneway_flag == '0':
                        way.oneway = False
                    elif oneway_flag == '-1':
                        way.oneway = True
                        way.is_reversed = True
                    elif oneway_flag in ['reversible', 'alternating']:
                        # todo: reversible, alternating: https://wiki.openstreetmap.org/wiki/Tag:oneway%3Dreversible
                        way.oneway = False
                    else:
                        print(f'  warning: new oneway flag {oneway_flag} detected at way {way.osm_way_id}')
                elif info.attrib['k'] == 'area':
                    way.area = info.attrib['v']
                elif info.attrib['k'] == 'motor_vehicle':
                    way.motor_vehicle = info.attrib['v']
                elif info.attrib['k'] == 'motorcar':
                    way.motorcar = info.attrib['v']
                elif info.attrib['k'] == 'service':
                    way.service = info.attrib['v']
                elif info.attrib['k'] == 'foot':
                    way.foot = info.attrib['v']
                elif info.attrib['k'] == 'bicycle':
                    way.bicycle = info.attrib['v']

        if way.highway:
            try:
                way.link_type_name = osm_highway_type_dict[way.highway]
                way.link_type = link_type_no_dict[way.link_type_name]
            except KeyError:
                continue

            allowable_agent_type_list = getAllowableAgentType(way)
            way.allowable_agent_type_list = list(set(allowable_agent_type_list).intersection(network_type_set))
            if len(way.allowable_agent_type_list) == 0:
                continue
            way.allowed_uses = ';'.join(way.allowable_agent_type_list)

            if way.ref_node_list[0] is way.ref_node_list[-1]:
                way.is_cycle = True

            if way.oneway is None:
                way.oneway = default_oneway_flag_dict[way.link_type_name]

            if way.lanes is None:
                if network.default_lanes:
                    way.lanes = network.default_lanes[way.link_type_name]
            if way.maxspeed is None:
                if network.default_speed:
                    way.maxspeed = network.default_speed[way.link_type_name]

            way_list.append(way)


    used_node_set = set()
    for way in way_list:
        way.ref_node_list[0].is_crossing = True
        way.ref_node_list[-1].is_crossing = True
        for node in way.ref_node_list[1:-1]:
            if node in used_node_set:
                node.is_crossing = True
            else:
                used_node_set.add(node)

    createLinks(network, way_list)


# todo: strict mode
def parseOSM(network, nodes, ways, strict_mode, network_type):
    parseNodes(network, nodes, strict_mode)
    parseWays(network, ways, network_type)


def removeUseless(network):
    remove_node_set = set()
    for node in network.node_set:
        if not (node.is_crossing and node.in_region):
            remove_node_set.add(node)
    for node in remove_node_set:
        network.node_set.remove(node)



def removeIsolated(network):
    isolated_node_set = set()
    isolated_link_set = set()

    for node in network.node_set:
        if len(node.incoming_link_list) + len(node.outgoing_link_list) == 0:
            node.is_isolated = True
            isolated_node_set.add(node)

    node_group_set = set()
    for link in network.link_set:
        new_group = (link.from_node, link.to_node)
        accessible_group_set = set()
        for node_group in node_group_set:
            if (link.from_node in node_group) or (link.to_node in node_group):
                accessible_group_set.add(node_group)

        for node_group in accessible_group_set:
            new_group += node_group
            node_group_set.remove(node_group)
        new_group = tuple(set(new_group))
        node_group_set.add(new_group)

    # select the set with the maximum number of nodes
    max_number_of_nodes = 0
    max_group = None
    for node_group in node_group_set:
        if len(node_group) > max_number_of_nodes:
            max_number_of_nodes = len(node_group)
            max_group = node_group

    for node_group in node_group_set:
        if node_group != max_group:
            for node in node_group:
                node.is_isolated = True
                isolated_node_set.add(node)

    for link in network.link_set:
        if link.from_node.is_isolated or link.to_node.is_isolated:
            link.is_isolated = True
            isolated_link_set.add(link)

    for node in isolated_node_set: network.node_set.remove(node)
    for link in isolated_link_set: network.link_set.remove(link)


def assignNewID(network):
    number_of_nodes = 0
    number_of_links = 0

    # ensure node_id and link_id keep unchanged between different runs
    node_list = sorted(network.node_set, key=lambda x: x.node_no)
    link_list = sorted(network.link_set, key=lambda x: x.link_no)

    for node in node_list:
        node.node_id = number_of_nodes
        network.node_id_to_node_dict[node.node_id] = node
        number_of_nodes += 1

    for link in link_list:
        link.link_id = number_of_links
        number_of_links += 1


def getValidNetworkType(network_type):
    if isinstance(network_type,str):
        network_type_temp = (network_type,)
    else:
        network_type_temp = network_type

    network_type_valid = []
    network_type_all = list(filters.keys())
    for net_type in network_type_temp:
        if net_type not in network_type_all:
            print(f'network type \'{net_type}\' does not belong to {network_type_all}, it will be skipped')
        else:
            network_type_valid.append(net_type)
    return network_type_valid


def getNetFromOSMFile(osm_filename='map.osm', network_type=('auto',), strict_mode=True, remove_isolated=True, simplify=True,
                      int_buffer=default_int_buffer, bbox=None, default_lanes=False, default_speed=False):
    """
    :param osm_filename:
    :param network_type:
    :param strict_mode:
    :param remove_isolated:
    :param simplify:
    :param int_buffer:
    :param bbox: (minlat, minlon, maxlat, maxlon)
    :param default_lanes: True; False; Dict
    :param default_speed: True; False; Dict
    :return: a network instance
    """
    # build network
    network = Network()

    if default_lanes:
        if isinstance(default_lanes,bool):
            network.default_lanes = default_lanes_dict
        elif isinstance(default_lanes,dict):
            network.default_lanes = default_lanes
            for link_type_name, lanes in default_lanes_dict.items():
                if link_type_name not in default_lanes.keys():
                    network.default_lanes[link_type_name] = lanes
        else:
            print('unsupported type for argument default_lanes, has set it to default value: False')
    if default_speed:
        if isinstance(default_speed,bool):
            network.default_speed = default_speed_dict
        elif isinstance(default_speed,dict):
            network.default_speed = default_speed
            for link_type_name, speed in default_speed_dict.items():
                if link_type_name not in default_speed.keys():
                    network.default_speed[link_type_name] = speed
        else:
            print('unsupported type for argument default_lanes, has set it to default value: False')


    bounds, nodes, ways = readXMLFile(osm_filename)
    if bbox is None:
        network.minlat, network.minlon, network.maxlat, network.maxlon = bounds['minlat'], bounds['minlon'], bounds['maxlat'], bounds['maxlon']
    else:
        network.minlat, network.minlon, network.maxlat, network.maxlon = bbox

    network_type_valid = getValidNetworkType(network_type)
    parseOSM(network, nodes, ways, strict_mode, network_type_valid)
    removeUseless(network)

    # remove isolated nodes and links
    if remove_isolated:
        removeIsolated(network)

    # merge adjacent links at two-degree nodes
    if simplify:
        simplifyNetwork(network)

    # identify complex intersections which contains multiple nodes
    identifyComplexIntersections(network,int_buffer)

    # assign new ids starting from 0 to nodes and links
    assignNewID(network)

    return network


def getNetFromCSV(folder=''):
    node_data, link_data = readCSVFile(folder)
    network = Network()

    for i in range(len(node_data)):
        node = Node()
        node.node_id = node_data.loc[i, 'node_id']
        osm_node_id = node_data.loc[i, 'osm_node_id']
        if osm_node_id == osm_node_id: node.osm_node_id = int(osm_node_id)
        osm_highway = node_data.loc[i, 'osm_highway']
        if osm_highway == osm_highway: node.osm_highway = osm_highway
        node.x_coord = node_data.loc[i, 'x_coord']
        node.y_coord = node_data.loc[i, 'y_coord']
        main_node_id = node_data.loc[i, 'main_node_id']
        if main_node_id == main_node_id: node.main_node_id = int(main_node_id)
        node.ctrl_type = node_data.loc[i, 'ctrl_type']
        network.node_set.add(node)

        network.node_id_to_node_dict[node.node_id] = node

    for i in range(len(link_data)):
        link = Link()
        name = link_data.loc[i, 'name']
        if name == name: link.name = name
        link.link_id = link_data.loc[i, 'link_id']
        link.osm_way_id = link_data.loc[i, 'osm_way_id']
        link.from_node = network.node_id_to_node_dict[link_data.loc[i, 'from_node_id']]
        link.to_node = network.node_id_to_node_dict[link_data.loc[i, 'to_node_id']]
        link.length = link_data.loc[i, 'length']
        lanes = link_data.loc[i, 'lanes']
        if lanes == lanes: link.lanes = lanes
        free_speed = link_data.loc[i, 'free_speed']
        if free_speed == free_speed: link.free_speed = free_speed
        link.link_type_name = link_data.loc[i, 'link_type_name']
        link.link_type = link_data.loc[i, 'link_type']
        link.geometry_str = link_data.loc[i, 'geometry']

        network.link_set.add(link)

        link.from_node.outgoing_link_list.append(link)
        link.to_node.incoming_link_list.append(link)

    return network




import xml.etree.cElementTree as ET
import pandas as pd
import os
import locale


def getBounds(element):
    try:
        minlat = float(element.attrib['minlat'])
    except KeyError:
        minlat = -90.0
    try:
        minlon = float(element.attrib['minlon'])
    except KeyError:
        minlon = -180.0
    try:
        maxlat = float(element.attrib['maxlat'])
    except KeyError:
        maxlat = 90.0
    try:
        maxlon = float(element.attrib['maxlon'])
    except KeyError:
        maxlon = 180.0
    return {'minlat':minlat, 'minlon':minlon, 'maxlat':maxlat, 'maxlon':maxlon}


def readXMLFile(osm_filename='map.osm'):
    # print(f'Loaing Network Data from ({osm_filename})')
    osmtree = ET.ElementTree(file=osm_filename)

    bounds = {'minlat':-90.0, 'minlon':-180.0, 'maxlat':90.0, 'maxlon':180.0}
    nodes = []
    ways = []

    osmnet = osmtree.getroot()
    for element in osmnet:
        if element.tag == 'bounds':
            bounds = getBounds(element)
        elif element.tag == 'node':
            nodes.append(element)
        elif element.tag == 'way':
            ways.append(element)

    return bounds, nodes, ways


def readCSVFile(folder):
    if folder:
        node_filepath = os.path.join(folder,'node.csv')
        link_filepath = os.path.join(folder,'link.csv')
    else:
        node_filepath = 'node.csv'
        link_filepath = 'link.csv'

    local_encoding = locale.getdefaultlocale()
    try:
        node_data = pd.read_csv(node_filepath)
    except UnicodeDecodeError:
        node_data = pd.read_csv(node_filepath,encoding=local_encoding[1])
    try:
        link_data = pd.read_csv(link_filepath)
    except UnicodeDecodeError:
        link_data = pd.read_csv(link_filepath, encoding=local_encoding[1])

    return node_data, link_data


if __name__ == '__main__':
    readXMLFile()



# Note: we adopt the filter logic from osmnx (https://github.com/gboeing/osmnx)
# exclude links with tag attributes in the filters
filters = {}


filters['auto'] = {'area':['yes'],
                   'highway':['cycleway','footway','path','pedestrian','steps','track','corridor','elevator','escalator',
                              'proposed','construction','bridleway','abandoned','platform','raceway','service'],
                   'motor_vehicle':['no'],
                   'motorcar':['no'],
                   'access':['private'],
                   'service':['parking','parking_aisle','driveway','private','emergency_access']
                   }

filters['bike'] = {'area':['yes'],
                   'highway':['footway','steps','corridor','elevator','escalator','motor','proposed','construction','abandoned','platform','raceway'],
                   'bicycle':['no'],
                   'service':['private'],
                   'access':['private']
                   }

filters['walk'] = {'area':['yes'],
                   'highway':['cycleway','motor','proposed','construction','abandoned','platform','raceway'],
                   'foot':['no'],
                   'service':['private'],
                   'access':['private']
                   }

def getAllowableAgentType(way):
    allowable_agent_type_list = []
    for agent_type, m_filter in filters.items():
        allowed = True
        for tag, exclude_list in m_filter.items():
            if eval(f'way.{tag} in exclude_list'):
                allowed = False
                break

        if allowed: allowable_agent_type_list.append(agent_type)
    return allowable_agent_type_list



from .util import *


class Segment:
    def __init__(self):
        self.link = None
        self.start_lr = 0.0
        self.end_lr = 0.0
        self.l_lanes_added = 0
        self.r_lanes_added = 0


class Node:
    def __init__(self):
        self.name = None
        self.node_id = 0
        self.osm_node_id = None
        self.node_no = 0
        self.x_coord = 0.0
        self.y_coord = 0.0
        self.main_node_id = None
        self.osm_highway = ''
        self.node_type = ''
        self.ctrl_type = 0
        self.in_region = True
        self.is_crossing = False
        self.is_isolated = False
        self.valid = True
        self.incoming_link_list = []
        self.outgoing_link_list = []



class Link:
    def __init__(self):
        self.link_id = 0
        self.osm_way_id = 0
        self.link_no = 0
        self.name = ''
        self.link_type_name = ''
        self.link_type = 0
        self.from_node = None
        self.to_node = None
        self.lanes = None
        self.lanes_list = []
        self.free_speed = None
        self.length = 0.0
        self.allowed_uses = ''
        self.geometry_point_list = []
        self.geometry_str = ''
        self.is_isolated = False
        self.valid = True
        self.ob_comb_link = None
        self.lanes_change_point_list = []

    def calculateLength(self):
        for i in range(len(self.geometry_point_list)-1):
            from_point = self.geometry_point_list[i]
            to_point = self.geometry_point_list[i+1]
            self.length += getDistanceFromCoord(from_point[0], from_point[1], to_point[0], to_point[1]) * 1000

    def getGeometryStr(self):
        geometry_str = 'LINESTRING (' + f'{self.geometry_point_list[0][0]} {self.geometry_point_list[0][1]}'
        for point in self.geometry_point_list[1:]: geometry_str += f', {point[0]} {point[1]}'
        geometry_str += ')'
        self.geometry_str = geometry_str


class Way:
    def __init__(self):
        self.osm_way_id = 0
        self.highway = ''
        self.link_type_name = ''
        self.link_type = 0
        self.name = ''
        self.lanes = None
        self.forward_lanes = None
        self.backward_lanes = None
        self.maxspeed = None
        self.oneway = None

        self.area = None
        self.motor_vehicle = None
        self.motorcar = None
        self.service = None
        self.access = None
        self.foot = None
        self.bicycle = None
        self.allowable_agent_type_list = []
        self.allowed_uses = ''

        self.is_reversed = False
        self.is_cycle = False
        self.ref_node_list = []
        self.number_of_segments = 0
        self.segment_node_list = []         # ref node sequence for each segment

    def getNodeListForSegments(self):
        number_of_ref_nodes = len(self.ref_node_list)
        last_idx = 0
        idx = 0

        while True:
            m_segment_node_list = [self.ref_node_list[last_idx]]
            for idx in range(last_idx+1,number_of_ref_nodes):
                ref_node = self.ref_node_list[idx]
                m_segment_node_list.append(ref_node)
                if ref_node.is_crossing:
                    last_idx = idx
                    break
            self.segment_node_list.append(m_segment_node_list)
            self.number_of_segments += 1
            if idx == number_of_ref_nodes-1: break


class Network:
    def __init__(self):
        self.minlat = 0.0
        self.minlon = 0.0
        self.maxlat = 0.0
        self.maxlon = 0.0

        self.created_nodes = 0
        self.created_links = 0

        self.default_lanes = False
        self.default_speed = False

        self.simplified = False
        self.complex_intersection_identified = False
        self.consolidated = False
        self.new_id = True

        self.node_set = set()
        self.link_set = set()
        self.segment_set = set()
        self.complex_intersection_set = set()

        self.osm_node_id_to_node_dict = {}
        self.node_id_to_node_dict = {}



from .classes import *
import pandas as pd


def consolidateComplexIntersections_old(network,external_file=None):
    network.consolidated = True

    max_node_id = 0
    node_group_dict = {}
    for node in network.node_set:
        if node.node_id > max_node_id:
            max_node_id = node.node_id
        if node.main_node_id is not None:
            if node.main_node_id in node_group_dict.keys():
                node_group_dict[node.main_node_id].append(node)
            else:
                node_group_dict[node.main_node_id] = [node]

    current_node_id = max_node_id + 1
    removal_node_set = set()
    removal_link_set = set()

    if external_file is None:
        complex_intersection_set = network.complex_intersection_set
    else:
        complex_intersection_set = set()

        complex_int_data = pd.read_csv(external_file)
        for i in range(len(complex_int_data)):
            group_id = complex_int_data.loc[i,'group_id']
            node_id_str = complex_int_data.loc[i,'node_id']
            node_id_str_list = node_id_str.split(';')

            try:
                node_group_list = [network.node_id_to_node_dict[int(node_id_str)] for node_id_str in node_id_str_list]
            except KeyError as e:
                print(f'cannot find node {e.args[0]} in the network when consolidating intersection group {group_id}, this group will keep unchanged')
                continue
            node_group = tuple(node_group_list)
            complex_intersection_set.add(node_group)

    for node_group in complex_intersection_set:
        new_node = Node()
        new_node.node_id = current_node_id
        x_coord_sum = 0
        y_coord_sum = 0
        osm_node_id_list = []

        for node in node_group:
            node.valid = False
            removal_node_set.add(node)

            x_coord_sum += node.x_coord
            y_coord_sum += node.y_coord
            osm_node_id_list.append(str(node.osm_node_id))

            for link in node.incoming_link_list:
                if link.from_node in node_group:
                    link.valid = False
                    removal_link_set.add(link)
                else:
                    link.to_node = new_node
            for link in node.outgoing_link_list:
                if link.to_node in node_group:
                    link.valid = False
                    removal_link_set.add(link)
                else:
                    link.from_node = new_node

            new_node.osm_highway = node.osm_highway

        new_node.x_coord = x_coord_sum / len(node_group)
        new_node.y_coord = y_coord_sum / len(node_group)
        new_node.osm_node_id = ';'.join(osm_node_id_list)

        network.node_set.add(new_node)
        current_node_id += 1

    for node in removal_node_set: network.node_set.remove(node)
    for link in removal_link_set: network.link_set.remove(link)



def consolidateComplexIntersections(network):
    max_node_id = 0
    node_group_dict = {}
    for node in network.node_set:
        if node.node_id > max_node_id:
            max_node_id = node.node_id
        if node.main_node_id is not None:
            if node.main_node_id in node_group_dict.keys():
                node_group_dict[node.main_node_id].append(node)
            else:
                node_group_dict[node.main_node_id] = [node]

    current_node_id = max_node_id + 1
    removal_node_set = set()
    removal_link_set = set()

    for main_node_id, node_group in node_group_dict.items():
        new_node = Node()
        new_node.node_id = current_node_id
        new_node.main_node_id = main_node_id
        new_node.ctrl_type = 1
        x_coord_sum = 0
        y_coord_sum = 0
        osm_node_id_list = []

        for node in node_group:
            node.valid = False
            removal_node_set.add(node)

            x_coord_sum += node.x_coord
            y_coord_sum += node.y_coord
            osm_node_id_list.append(str(node.osm_node_id))

            for link in node.incoming_link_list:
                if link.from_node in node_group:
                    link.valid = False
                    removal_link_set.add(link)
                else:
                    link.to_node = new_node
            for link in node.outgoing_link_list:
                if link.to_node in node_group:
                    link.valid = False
                    removal_link_set.add(link)
                else:
                    link.from_node = new_node

            new_node.osm_highway = node.osm_highway

        new_node.x_coord = x_coord_sum / len(node_group)
        new_node.y_coord = y_coord_sum / len(node_group)
        new_node.osm_node_id = ';'.join(osm_node_id_list)

        network.node_set.add(new_node)
        current_node_id += 1

    for node in removal_node_set: network.node_set.remove(node)
    for link in removal_link_set: network.link_set.remove(link)


import csv
import os


def outputNetToCSV(network, output_folder=''):
    if output_folder:
        if not os.path.exists(output_folder): os.mkdir(output_folder)
        node_filepath = os.path.join(output_folder,'node.csv')
        link_filepath = os.path.join(output_folder,'link.csv')
        segment_filepath = os.path.join(output_folder,'segment.csv')
    else:
        node_filepath = 'node.csv'
        link_filepath = 'link.csv'
        segment_filepath = 'segment.csv'

    # we use errors='ignore' to make our program compatible with some characters that cannot be encoded by the local encoding,
    # these characters will be discarded

    while True:
        try:
            outfile = open(node_filepath, 'w', newline='',errors='ignore')
            break
        except PermissionError:
            print('node.csv may be locked by other programs. please release it then press Enter to try again')
            input()
    writer = csv.writer(outfile)
    writer.writerow(['name', 'node_id', 'osm_node_id', 'osm_highway', 'zone_id', 'ctrl_type', 'node_type', 'activity_type',
                     'is_boundary', 'x_coord', 'y_coord', 'geometry','main_node_id'])
    node_list = sorted(network.node_set, key=lambda x: x.node_id)
    for node in node_list:
        line = [node.name, node.node_id, node.osm_node_id, node.osm_highway, '', node.ctrl_type, '', '', '', node.x_coord, node.y_coord,'',node.main_node_id]
        writer.writerow(line)
    outfile.close()

    while True:
        try:
            outfile = open(link_filepath, 'w', newline='',errors='ignore')
            break
        except PermissionError:
            print('link.csv may be locked by other programs. please release it then press Enter to try again')
            input()
    writer = csv.writer(outfile)
    writer.writerow(['name', 'link_id', 'osm_way_id', 'from_node_id', 'to_node_id', 'dir_flag', 'length', 'lanes',
                     'free_speed', 'capacity', 'link_type_name', 'link_type', 'geometry','allowed_uses'])
    link_list = sorted(network.link_set, key=lambda x: x.link_id)
    for link in link_list:
        line = [link.name, link.link_id, link.osm_way_id, link.from_node.node_id, link.to_node.node_id, '',link.length,
                link.lanes, link.free_speed, '', link.link_type_name, link.link_type, link.geometry_str, link.allowed_uses]
        writer.writerow(line)
    outfile.close()

    if network.simplified:
        while True:
            try:
                outfile = open(segment_filepath, 'w', newline='',errors='ignore')
                break
            except PermissionError:
                print('segment.csv may be locked by other programs. please release it then press Enter to try again')
                input()
        writer = csv.writer(outfile)
        writer.writerow(['segment_id','link_id','ref_node_id','start_lr','end_lr','capacity','free_speed','bike_facility',
                         'ped_facility','parking','allowed_uses','l_lanes_added','r_lanes_added'])
        segment_list = sorted(network.segment_set, key=lambda x: x.link.link_id)
        for segment_no, segment in enumerate(segment_list):
            line = [segment_no, segment.link.link_id,'',segment.start_lr,segment.end_lr,'','','','','','',segment.l_lanes_added,'']
            writer.writerow(line)
        outfile.close()


# -*- coding:utf-8 -*-
import setuptools

setuptools.setup(
    name='osm2gmns',
    version='0.0.4',
    author='Jiawei Lu, Xuesong Zhou',
    author_email='jiaweil9@asu.edu, xzhou74@asu.edu',
    url='https://github.com/jiawei92/OSM2GMNS',
    description='build network with gmns format using .osm file from openstreetmap',
    long_description=open('README_pypi.rst').read(),
    packages=['osm2gmns'],
    install_requires=[]
)
